from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM
import numpy as np

class BHMMTestModel(BernoulliHMM):
    def __init__(self, ctrl):
        BernoulliHMM.__init__(self, ctrl)

    def predict(self, X):
        """
        computes the the probs of classes for every feature vector
        x in the column seperatly.
        todo is not optimal (how can I do this???)
            i want a sequence to be estimated
            do this with function input formatter
            do this with output formatter

        :param X:
        :return:
        nd array (1D)
            n times the predictions
        """
        if len(X.shape) == 1:
            X = np.reshape(X, (-1, len(X)))
        T, _ = X.shape
        resnp = np.zeros((T, self.K), dtype=np.float64)
        for i, row in enumerate(X):
            row = np.reshape(row, (-1, len(row)))
            tmp = super().classify_multi(row, as_dict=False)
            resnp[i, :] = tmp
        return resnp

from hassbrain_algorithm.models.hmm.bhsmm import BHSMM
class BHSMMTestModel(BHSMM):
    def __init__(self, ctrl):
        BHSMM.__init__(self, ctrl)

    def predict(self, X):
        """
        computes the the probs of classes for every feature vector
        x in the column seperatly.
        todo is not optimal (how can I do this???)
            i want a sequence to be estimated
            do this with function input formatter
            do this with output formatter

        :param X:
        :return:
        nd array (1D)
            n times the predictions
        """
        if len(X.shape) == 1:
            X = np.reshape(X, (-1, len(X)))
        T, _ = X.shape
        resnp = np.zeros((T, self.K), dtype=np.float64)
        for i, row in enumerate(X):
            row = np.reshape(row, (-1, len(row)))
            tmp = super().classify_multi(row, as_dict=False)
            resnp[i, :] = tmp
        return resnp



from hassbrain_algorithm.controller import Controller, Dataset
from hassbrain_algorithm.datasets._dataset import DataRep
from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM

# model names
from scripts.test_model import BHMMTestModel
from scripts.test_model import BHSMMTestModel

BHMM = 'bhmm'
BHMMPC = 'bhmmpc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'
MCTADS = 'tads' # test activity distribution sampler


SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC20 = '0.3min'
SEC30 = '0.5min'

MIN1 = '1min'
MIN30 = '30min'
DK = Dataset.HASS

DN_KASTEREN = 'kasteren'
DN_HASS_TESTING2 = 'hass_testing2'
DN_HASS_CHRIS = 'hass_chris'
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
BASE_PATH = '/home/cmeier/code/data/thesis_results'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

MODE_TRAIN = 'train'
MODE_BENCH = 'bench'

#------------------------
# Configuration

MODEL_CLASS = BHMM
DATA_NAME = DN_HASS_CHRIS_FINAL
DATA_TYPE = DT_RAW
TIME_DIFF = MIN1
MODE = MODE_TRAIN
BEST_MODEL_SELECTION = False

#------------------------

DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

MODEL_NAME = 'model_' + MODEL_CLASS + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF
if BEST_MODEL_SELECTION:
    MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/' + MODEL_NAME
else:
    MODEL_FOLDER_PATH = BASE_PATH + '/' + DATA_NAME + '/models/' + MODEL_NAME

MODEL_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME +".joblib"
MD_LOSS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.loss.csv'
MD_LOSS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.loss.png'
MD_INFST_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.inferred_states.png'
MD_CONF_MAT_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.confusion_matrix.numpy'
MD_METRICS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.metrics.csv'
MD_CLASS_ACTS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.class_acts.csv'
MD_ACT_DUR_DISTS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.act_dur_dists.png'
MD_ACT_DUR_DISTS_DF_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.act_dur_dists.csv'
DATA_ACT_DUR_DISTS_IMG_FILE_PATH = MODEL_FOLDER_PATH + '/dataset.act_dur_dists.png'
DATA_ACT_DUR_DISTS_DF_FILE_PATH = MODEL_FOLDER_PATH + '/dataset.act_dur_dists.csv'
MD_FEATURE_IMP_PLT_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.feature_importance.png'

def main():
    ctrl = Controller()
    ctrl.load_dataset_from_file(DATASET_FILE_PATH)
    from scripts.test_model import BHSMMTestModel
    from hassbrain_algorithm.models.hmm.bhmm_hp import BernoulliHMM_HandcraftedPriors
    from hassbrain_algorithm.models.tads import TADS

    if MODEL_CLASS == BHMM:
        hmm_model = BHMMTestModel(ctrl)
        hmm_model.set_training_steps(10)
    elif MODEL_CLASS == BHSMM:
        hmm_model = BHSMMTestModel(ctrl)
        hmm_model.set_training_steps(50)
    elif MODEL_CLASS == BHMMPC:
        hmm_model = BernoulliHMM_HandcraftedPriors(ctrl)
    elif MODEL_CLASS == MCTADS:
        hmm_model = TADS(ctrl)
    else:
        raise ValueError

    ctrl.register_model(hmm_model, MODEL_NAME)

    ctrl.register_benchmark(MODEL_NAME)
    ctrl.init_model_on_dataset(MODEL_NAME)
    ctrl.register_loss_file_path(MD_LOSS_FILE_PATH, MODEL_NAME)
    ctrl.train_model(MODEL_NAME)
    ctrl.save_model(MODEL_FILE_PATH, MODEL_NAME)

    # bench the model
    params = {
       'metrics' : False,
       'act_dur_dist' : True,
        'feature_importance' : False
    }
    reports = ctrl.bench_models(**params)

    # save metrics
    #ctrl.save_df_metrics_to_file(MODEL_NAME, MD_METRICS_FILE_PATH)
    #ctrl.save_df_confusion(MODEL_NAME, MD_CONF_MAT_FILE_PATH)
    #ctrl.save_df_act_dur_dists(MODEL_NAME, MD_ACT_DUR_DISTS_DF_FILE_PATH,
    #                           DATA_ACT_DUR_DISTS_DF_FILE_PATH)
    #ctrl.save_df_class_accs(MODEL_NAME, MD_CLASS_ACTS_FILE_PATH)
    dict = {
        #'feature_importance' : MD_FEATURE_IMP_PLT_FILE_PATH,
        'train_loss' : MD_LOSS_IMG_FILE_PATH,
        #'inf_states' : MD_INFST_IMG_FILE_PATH,
        'act_dur' : MD_ACT_DUR_DISTS_IMG_FILE_PATH
    }
    ctrl.save_plots(MODEL_NAME, dict)

if __name__ == '__main__':
    print('dataset filepath : ', DATASET_FILE_PATH)
    print('model filepath: ', MODEL_FILE_PATH)
    main()
    print('finished')


# tutorial is from docs
# 'https://eli5.readthedocs.io/en/latest/blackbox/permutation_importance.html'
import ssm

from hassbrain_algorithm.controller import Controller, Dataset
from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM
"""

# feature importance
    how well does sensor x_i reduce the prediction error
        mean decrease impurity (is model specific therefore don't use)

        - permutation importance (MDA)
            feature is unimportant if permuting values leaves model error unchanged
            eli5

# PDP ICE plots
    ice
        how changes of a given featuer impacts predictions for set of observations

"""
# model names
BHMM = 'bhmm'
BHMMPC = 'bhmm_pc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'

SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC30 = '0.5min'

MIN1 = '1min'
MIN30 = '30min'
DK = Dataset.HASS

DN_KASTEREN = 'kasteren'
DN_HASS_TESTING2 = 'hass_testing2'
DN_HASS_CHRIS = 'hass_chris'
BASE_PATH = '/home/cmeier/code/data/thesis_results'
MD_BASE_PATH = '/home/cmeier/code/tmp'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

#------------------------
# Configuration

MODEL_CLASS = BHMM
DATA_NAME = DN_KASTEREN
DATA_TYPE = DT_RAW
TIME_DIFF = MIN1

#------------------------

DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

MODEL_NAME = 'model_' + MODEL_CLASS + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF
MODEL_FOLDER_PATH = MD_BASE_PATH + '/' + DATA_NAME + '/models/' + MODEL_NAME
MODEL_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME +".joblib"



"""
x bernoulli vector lenght k -> one hot encoded 2*K vector
y -> (probability of class)
"""

import numpy as np
path_to_config = '/home/cmeier/code/hassbrain_algorithm/hassbrain_algorithm/config.yaml'
X = [[1, 0, 1, 0],
     [1, 0, 1, 0],
     [1, 1, 1, 0],
     [1, 1, 1, 0],
     [1, 1, 1, 1],
     [1, 1, 1, 0]]

ypr = [0, 0, 1, 1, 2, 1]

#_-------------------------------------------------------------------------------------
#ctrl = Controller(path_to_config=path_to_config)
#ctrl.load_dataset_from_file(DATASET_FILE_PATH)
#from scripts.test_model import TestModel
#hmm_model = TestModel(ctrl)
##hmm_model = BernoulliHMM(ctrl)
#ctrl.load_dataset_from_file(DATASET_FILE_PATH)
#ctrl.load_model(MODEL_FILE_PATH, MODEL_NAME)
##ctrl.register_model(hmm_model, MODEL_NAME)
##ctrl.register_benchmark(MODEL_NAME)
##ctrl.init_model_on_dataset(MODEL_NAME)
##hmm_model.set_training_steps(100)
##ctrl.train_model(MODEL_NAME)
#
#
#
#
#
#model = ctrl._models[MODEL_NAME] # type: BernoulliHMM
#dataset = ctrl._dataset
#idxs = np.random.choice(len(dataset._test_all_x), size=10)
#tmp1 = dataset._test_all_x
#tmp2 = dataset._test_all_y
#train_z = dataset._test_all_x[idxs]
#train_x = dataset._test_all_y[idxs]



#_-------------------------------------------------------------------------------------
T = 20  # number of time bins
K = 4   # number of discrete states
D = 5   # dimension of the observations

KTrue = 3    # number of discrete states

hmm = ssm.HSMM(KTrue, D, observations="bernoulli")
z_true, x_true = hmm.sample(T)
def str_arr2boolean(arr):
    return arr == 'True'
#Fitting an HMM is simple.
class TestModel():
    def __init__(self, x_true):
        self.test_hmm = ssm.HMM(K, D, observations="bernoulli")
        self.test_hmm.fit(x_true)
        self.K = K

    def predict(self, X):
        X = str_arr2boolean(X)
        z_pred = self.test_hmm.most_likely_states(X)
        z_labeld_pred = np.zeros_like(z_pred, dtype=object)
        for i, item in enumerate(z_pred):
            z_labeld_pred[i] = str('z' + str(z_pred[i]))
        return z_labeld_pred
        #return z_pred

    def predict_proba(self, X):
        """Predict class probabilities for X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        Raises
        ------
        AttributeError
            If the ``loss`` does not support probabilities.

        Returns
        -------
        p : array, shape (n_samples, n_classes)
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute `classes_`.
        """
        if len(X.shape) == 1:
            X = np.reshape(X, (-1, len(X)))
        T, _ = X.shape
        resnp = np.zeros((T, self.test_hmm.K), dtype=np.float64)
        for i, row in enumerate(X):
            row = np.reshape(row, (-1, len(row)))
            bool_row = num2boolean(row)
            tmp = self.test_hmm.filter(bool_row)
            resnp[i, :] = tmp
        return resnp

def num2boolean(arr):
    res = arr.astype(bool)
    return res

train_x = x_true
train_z = z_true
model = TestModel(x_true)
class_names = ['z0', 'z1', 'z2', 'z3', 'z4']#, 'z5', 'z6', 'z7']
feature_names = ['f0', 'f1', 'f2', 'f3', 'f4']#, 'f5', 'f6']#, 'f6', 'f7', 'f8', 'f9', 'f10', 'f11', 'f12']

from skater.model import InMemoryModel
from skater.core.explanations import Interpretation
from matplotlib.pyplot import Figure
#feature_names = hmm_model.get_obs_lbl_lst()
#class_names = hmm_model.get_state_lbl_lst()

def boolean_arr2str(arr):
    res = arr.astype(str)
    return res


cat_idxs = [i for i in range(len(class_names))]
categorical_names = {}
for i in cat_idxs:
    categorical_names[i] = {}
    categorical_names[i][0] = "off"
    categorical_names[i][1] = "on"

from skater.core.local_interpretation.lime.lime_tabular import LimeTabularExplainer
exp = LimeTabularExplainer(
    train_x,
    mode='classification',
    training_labels=train_z,
    feature_names=feature_names,
    categorical_features=cat_idxs,
    categorical_names=categorical_names,
    class_names=class_names)

#fig = exp.explain_instance(train_x[0], model.predict_proba).as_pyplot_figure()
#fig.show()
x = train_x[0]
lst = exp.explain_instance(x, model.predict_proba).as_list()
print(lst)



import numpy as np
import pandas as pd
import dask
from pyadlml.dataset import START_TIME, END_TIME, TIME, \
    TIME, NAME, VAL, DEVICE
from pyadlml.util import get_parallel, get_npartitions

"""
    df_devices:
        also referred to as rep1
        a lot of data is found in this format.
        Exc: 
           | time      | device    | state
        ------------------------------------
         0  | timestamp | dev_name  |   1

    rep2:
        is used to calculate statistics for devices more easily
        and has lowest footprint in storage. Most of the computation 
        is done using this format
           | start_time | end_time   | device    
        ----------------------------------------
         0 | timestamp   | timestamp | dev_name  
 
 """

def _create_devices(dev_list, index=None):
    """
    creates an empty device dataframe
    """
    if index is not None:
        return pd.DataFrame(columns=dev_list, index=index)
    else:
        return pd.DataFrame(columns=dev_list)



def _check_devices_sequ_order(df):
    """
    iterate pairwise through each select device an check if the 
    sequential order in time is inconsistent

    Parameters
    ----------
    df : pd.DataFrame
        device representation 1  with columns [start_time, end_time, devices]
    """
    dev_list = df['device'].unique()
    no_errors = True
    for dev in dev_list:
        df_d = df[df['device'] == dev]
        for i in range(1,len(df_d)):
            st_j = df_d.iloc[i-1].start_time
            et_j = df_d.iloc[i-1].end_time
            st_i = df_d.iloc[i].start_time
            et_i = df_d.iloc[i].end_time
            # if the sequential order is violated return false
            if not (st_j < et_j) or not (et_j < st_i):
                print('~'*50)
                if (st_j >= et_j):                    
                    #raise ValueError('{}; st: {} >= et: {} |\n {} '.format(i-1, st_j, et_j, df_d.iloc[i-1]))
                    print('{}; st: {} >= et: {} |\n {} '.format(i-1, st_j, et_j, df_d.iloc[i-1]))
                if (et_j >= st_i):                    
                    #raise ValueError('{},{}; et: {} >= st: {} |\n{}\n\n{}'.format(i-1,i, et_j, st_i, df_d.iloc[i-1], df_d.iloc[i]))
                    print('{},{}; et: {} >= st: {} |\n{}\n\n{}'.format(i-1,i, et_j, st_i, df_d.iloc[i-1], df_d.iloc[i]))
                no_errors = False
    return no_errors



def _is_dev_rep2(df):
    """
    """
    if not START_TIME in df.columns or not END_TIME in df.columns \
    or not DEVICE in df.columns or len(df.columns) != 3:
        return False
    return True

def _is_dev_rep1(df):
    """
    """
    if DEVICE in df.columns and VAL in df.columns \
    and TIME in df.columns and len(df.columns) == 3:
        return True
    return False

def device_rep1_2_rep2(df_rep1, drop=False):
    """ transforms a device representation 1 into 2
    Parameters
    ----------
    df_rep1 : pd.DataFrame
        rep1: col (time, device, val)
        example row: [2008-02-25 00:20:14, Freezer, False]
    drop : Boolean
        Indicates whether rows that have no starting/end timestamp should be dropped or the
        missing counterpart should be added with the first/last timestamp
    Returns
    -------
    df : pd.DataFrame
        rep: columns are (start time, end_time, device)
        example row: [2008-02-25 00:20:14, 2008-02-25 00:22:14, Freezer]         
    or 
    df, lst
    """
    df = df_rep1.copy().reset_index(drop=True)
    df = df.sort_values(TIME)
    df.loc[:,'ones'] = 1
    
    rows_changed = 0
    syn_acts = []
    if drop:
        to_delete_idx = []
        for dev in df['device'].unique():
            df_dev = df[df['device'] == dev]
            first_row = df_dev.iloc[0].copy()
            last_row = df_dev.iloc[len(df_dev)-1].copy()
            if not first_row['val']:
                to_delete_idx.append(first_row.name)
            if last_row['val']:
                to_delete_idx.append(last_row.name)
        df = df.drop(to_delete_idx)
        rows_changed = -len(to_delete_idx)
    else:
        # add values to things that are false
        first_timestamp = df['time'].iloc[0]
        last_timestamp = df['time'].iloc[len(df)-1]
        eps = pd.Timedelta('1ns')
        for dev in df['device'].unique():
            df_dev = df[df['device'] == dev]
            first_row = df_dev.iloc[0].copy()
            last_row = df_dev.iloc[len(df_dev)-1].copy()
            if not first_row['val']:
                first_row['val'] = True
                first_row['time'] = first_timestamp + eps
                syn_acts.append(first_row)
                df = df.append(first_row, ignore_index=True)
            if last_row['val']:
                last_row['val'] = False
                last_row['time'] = last_timestamp - eps
                syn_acts.append(last_row)
                df = df.append(last_row, ignore_index=True)
            eps += pd.Timedelta('1ns')
        rows_changed = len(syn_acts)
            
    df = df.reset_index(drop=True).sort_values(TIME)
    
    # seperate the 0to1 and 1to0 device changes
    df.loc[:,VAL] = df[VAL].astype(bool)
    df_start = df[df[VAL]] 
    df_end = df[~df[VAL]]    
    df_end = df_end.rename(columns={TIME: END_TIME})
    df_start = df_start.rename(columns={TIME: START_TIME})

    # ordered in time to index them and make a correspondence
    df_end.loc[:,'pairs'] = df_end.groupby([DEVICE])['ones'].apply(lambda x: x.cumsum())
    df_start.loc[:,'pairs'] = df_start.groupby([DEVICE])['ones'].apply(lambda x: x.cumsum())        
        
    
    df = pd.merge(df_start, df_end, on=['pairs', DEVICE])
    df = df.sort_values(START_TIME)
    
    # sanity checks        
    diff = int((len(df_rep1)+rows_changed)/2) - len(df)
    assert diff == 0, 'input {} - {} == {} result. Somewhere two following events of the \
        #        same device had the same starting point and end point'.format(int(len(df_rep1)/2), len(df), diff)
    
    if drop:
        return df[[START_TIME, END_TIME, DEVICE]]
    else:
        return df[[START_TIME, END_TIME, DEVICE]], syn_acts

def device_rep2_2_rep1(df_rep2):
    """
    Parameters
    ----------
    df_rep2 : pd.DataFrame
        rep2: columns (start time, end_time, device)

    Returns
    -------
    df : (pd.DataFrame)
        rep1: columns are (time, value, device)
        example row: [2008-02-25 00:20:14, Freezer, False]
    """
    # copy devices to new dfs 
    # one with all values but start time and other way around
    df_start = df_rep2.copy().loc[:, df_rep2.columns != END_TIME]
    df_end = df_rep2.copy().loc[:, df_rep2.columns != START_TIME]

    # set values at the end time to zero because this is the time a device turns off
    df_start[VAL] = True
    df_end[VAL] = False

    # rename column 'End Time' and 'Start Time' to 'Time'
    df_start.rename(columns={START_TIME: TIME}, inplace=True)
    df_end.rename(columns={END_TIME: TIME}, inplace=True)

    df = pd.concat([df_end, df_start]).sort_values(TIME)
    df = df.reset_index(drop=True)
    return df

def correct_device_rep1_ts_duplicates(df):
    """
    remove devices that went on and off at the same time. And add a microsecond
    to devices that trigger on the same time
    Parameters
    ----------
    df : pd.DataFrame
        Devices in representation 1; columns [time, device, value]
    """
    eps = pd.Timedelta('10ms')
    
    try:
        df[TIME]
    except KeyError:
        df = df.reset_index()
        
    # remove device if it went on and off at the same time    
    dup_mask = df.duplicated(subset=[TIME, DEVICE], keep=False)    
    df = df[~dup_mask]
    
    df = df.reset_index(drop=True)
    
    
    dup_mask = df.duplicated(subset=[TIME], keep=False)
    duplicates = df[dup_mask]    
    uniques = df[~dup_mask]
    
    # for every pair of duplicates add a millisecond on the second one
    duplicates = duplicates.reset_index(drop=True)
    sp = duplicates['time'] + eps   
    mask_p = (duplicates.index % 2 == 0) 
    
    duplicates['time'] = duplicates['time'].where(mask_p, sp)
    
    # concatenate and sort the dataframe 
    uniques = uniques.set_index(TIME)
    duplicates = duplicates.set_index(TIME)
    df = pd.concat([duplicates, uniques], sort=True)
    
    # set the time as index again
    df = df.sort_values(TIME).reset_index(drop=False)
    
    return df
    
def _has_timestamp_duplicates(df):
    """ check whether there are duplicates in timestamp present
    Parameters
    ----------
    df : pd.DataFrame
        data frame representation 1: [time, device, val]
    """
    df = df.copy()
    try:
        dup_mask = df.duplicated(subset=[TIME], keep=False)
    except KeyError:
        df = df.reset_index()
        dup_mask = df.duplicated(subset=[TIME], keep=False)
    return dup_mask.sum() > 0

def correct_devices(df):
    """
    Parameters
    ----------
    df : pd.DataFrame
        either in device representation 1 or 2
    Returns
    -------
    cor_rep1 : pd.DataFrame
    """
    df = df.copy()
    df = df.drop_duplicates()        
    
    # bring in correct representation
    if _is_dev_rep2(df):
        cor_rep1 = device_rep2_2_rep1(df)        
    elif _is_dev_rep1(df):
        cor_rep1 = df
    else:
        raise ValueError
    cor_rep1 = cor_rep1.sort_values(by='time').reset_index(drop=True)

    # correct timestamp duplicates
    while _has_timestamp_duplicates(cor_rep1):
        cor_rep1 = correct_device_rep1_ts_duplicates(cor_rep1)

    # correct on off incosistency
    if not is_on_off_consistent(cor_rep1):
        cor_rep1 = correct_on_off_inconsistency(cor_rep1)

    assert not _has_timestamp_duplicates(cor_rep1)
    assert is_on_off_consistent(cor_rep1)
   
    return cor_rep1

def is_on_off_consistent(df):
    """ devices can only go on after they are off and vice versa. check if this is true
        for every column
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe in representation 1.
    """
    # check if the alternating pattern of devices holds
    def func(df, dev):
        """ returns whether there are rows that are different to an alternating pattern
        """
        df_dev = df[df['device'] == dev].copy().sort_values(by='time').reset_index(drop=True)
        first_val = df_dev['val'].iloc[0]
        if first_val:            
            mask = np.zeros((len(df_dev)), dtype=bool)
            mask[::2] = True
        else:
            mask = np.ones((len(df_dev)), dtype=bool)
            mask[::2] = False 
    
        return not (df_dev['val'] ^ mask).sum() > 0

    lazy_results = []
    for dev in df['device'].unique():
        res = dask.delayed(func)(df, dev)
        lazy_results.append(res)

    results = np.array(list(dask.compute(*lazy_results)))
    return results.any() 



from dask import delayed
def correct_on_off_inconsistency(df):
    """ 
    has multiple strategies for solving the patterns:
    e.g if the preceeding value is on and the same value is occuring now delete now
    Parameters
    ----------
    df : pd.DataFrame
        device representation 3
        
    Returns
    -------
    df : pd.DataFrame
        device representation 3
    """
    
    def get_inconsistent_series(df_dev):            
        # create alternating mask depending on first value            
        first_val = df_dev['val'].iloc[0]
        if first_val:            
            mask = np.zeros((len(df_dev)), dtype=bool)
            mask[::2] = True
        else:
            mask = np.ones((len(df_dev)), dtype=bool)
            mask[::2] = False 
    
        df_dev['diff'] = df_dev['val'] ^ mask
        if df_dev['diff'].sum() > 0:
            val = True
        else:
            val = False
        return (val, df_dev[['time', 'device', 'val']])
        
        
    def get_inconsistent_parts(df, dev):
        df_dev = df[df['device'] == dev].copy().sort_values(by='time').reset_index()
        return get_inconsistent_series(df_dev)
                    
    def correct_part(df_dev):
        """ get index of rows where the previous value was the same and delete row
        """
        
        df_dev['same_prec'] = ~(df_dev['val'].shift(1) ^ df_dev['val'])
        df_dev.loc[0, 'same_prec'] = False # correct shift artefact
        indices = list(df_dev[df_dev['same_prec']].index)
        df_dev = df_dev.drop(indices, axis=0)
        return df_dev[['time', 'device', 'val']]
                    
    def split_incs(dev_df_list):
        li, lc = [], []
        for incon, part in dev_df_list:
            if incon: li.append(part)
            else: lc.append(part)
        return lc, li
    
    # create list of tuples e.g [(True, df_dev1), (False, df_dev2), ...]
    dev_list = df['device'].unique()
    dev_df_list = []
    for devs in dev_list:
        dev_df_list.append(delayed(get_inconsistent_parts)(df, devs))
    cons_lst, incons_lst = delayed(split_incs)(dev_df_list).compute()
    
    
    tmp = []
    for part in incons_lst:
        corr_part = delayed(correct_part)(part)
        tmp.append(delayed(get_inconsistent_series)(corr_part))
    tmp = delayed(lambda x: x)(tmp).compute()
    
    # check if all series is on off consistent
    for incon, part in tmp:
        assert not incon
        cons_lst.append(part)
    
    return pd.concat(cons_lst, ignore_index=True)\
            .sort_values(by='time').reset_index(drop=True)


import numpy as np
import matplotlib.pyplot as plt


labels = ['ts acc', 'class acc']
hmm = [22, 17]
hsmm = [23, 20]
bhmm = [35, 27]
bhsmm = [41, 30]
pcbhmm = [32, 15]
pcbhsmm = [20, 10]

x = np.arange(len(labels))  # the label locations
bar_count = 3
width = 0.10  # the width of the bars
offset = width * bar_count - 0.025
step = width + 0.01
fig, ax = plt.subplots()
color=['black', 'grey', 'lightgrey', 'dimgray', 'darkgray', 'darkgray']
bar0 = ax.bar(x - offset + 0*step, hmm,     width, color=color[0], label='hmm')
bar1 = ax.bar(x - offset + 1*step, hsmm,    width, color=color[1], label='hsmm')
bar2 = ax.bar(x - offset + 2*step, bhmm,    width, color=color[2], label='bhmm')
bar3 = ax.bar(x - offset + 3*step, bhsmm,   width, color=color[3], label='bhsmm')
bar4 = ax.bar(x - offset + 4*step, pcbhmm,  width, color=color[4], label='pcbhmm')
bar5 = ax.bar(x - offset + 5*step, pcbhsmm,  width, color=color[5], label='pcbhsmm')

# Add some text for labels, title and custom x-axis tick labels, etc.
ax.set_ylabel('accuracy')
ax.set_xticks(x)
ax.set_xticklabels(labels)
legend = ax.legend(loc='middle')
plt.show()



import math
from hassbrain_algorithm.datasets._dataset import _Dataset
from hassbrain_algorithm.models._model import Model
import numpy as np

class TADS(Model):

    def __init__(self, controller):
        self._activity_dur_dist = None
        self._cum_act_dur_dist = None
        Model.__init__(self, "test", controller)

    def __str__(self):
        s = ""
        return s

    def _model_init(self, dataset : _Dataset):
        import pandas as pd
        tmp = dataset.get_rel_act_durations() # type: pd.Dataframe
        hm = self._state_lbl_hashmap
        tmp.rename(columns=hm,
                 inplace=True)
        self._activity_dur_dist = tmp.copy()
        self._cum_act_dur_dist = tmp.cumsum(axis=1)

    def get_act_dur_dist(self):
        rev_hm = self._state_lbl_rev_hashmap
        assert self._activity_dur_dist is not None
        tmp = self._activity_dur_dist.rename(columns=rev_hm,
                 inplace=False)
        return tmp

    def save_visualization(self, path_to_file):
        """
        save a visualization of the model to the given filepath
        :param path_to_file:
        :return:
        """
        raise NotImplementedError

    def predict_state_sequence(self, test_y):
        l, _ = test_y.shape
        pred_z, _ = self._sample(l)
        return pred_z


    def predict_obs_sequence(self, test_y):
        raise NotImplemented

    def can_predict_next_obs(self):
        return False

    def _draw_sample(self):
        s = np.random.uniform(low=0.0, high=1.0)
        for i, val in enumerate(self._cum_act_dur_dist.iloc[0]):
            if s < val:
                return int(i)
        raise ValueError

    def _sample(self, n, obs_seq=None):
        res = np.zeros((n), dtype=np.int64)
        for i in range(n):
            res[i] = self._draw_sample()
        z = res
        obs = None
        return z, obs

    def can_predict_prob_devices(self):
        return False

    def _train(self, dataset):
        return None

    def _classify(self, obs_seq):
        """
        get the most probable state/activity
        :param obs_seq:
        :return:
        """
        assert len(obs_seq[0]) == self._hmm.D
        state_seq = self._hmm.most_likely_states(obs_seq)
        pred_state = state_seq[-1:][0]
        return pred_state

    def _classify_multi(self, obs_seq):
        """
        computes the last omega slice of viterbi which is
        equivalent to
        :param obs_seq:
        :return:
            numpy array with each index being the id for the label
        """
        tmp = self._hmm.filter(obs_seq)
        last_alpha = tmp[-1:][0]


        assert math.isclose(last_alpha.sum(), 1.)

        K = self._hmm.K
        return last_alpha



from pyadlml.dataset.stats.discrete import contingency_table_01, cross_correlation
from pyadlml.dataset.plot.util import annotate_heatmap, heatmap, heatmap_square

import matplotlib.pyplot as plt 
import matplotlib
import pandas as pd 
import numpy as np

def heatmap_contingency_01(X, y, rep='', z_scale=None, figsize=(16,12)):
    """ plots the contingency between features and labels of data
    Parameters
    ----------
    X: pd.DataFrame
        one of the representation raw, lastfired, changepoint
    y: pd.DataFrame
        a series of labels
    rep: string
        the name of the representation to add to the title
    """
    
    cbarlabel = 'counts'
    title = "On/Off contingency for representation " + rep

    df_con = contingency_table_01(X, y)
    
    vals = df_con.values.T
    acts = df_con.columns
    devs = list(df_con.index)
    
    # format x labels
    for i in range(0,len(devs)):
        if i % 2 == 0:
            tmp = devs[i][:-3]
            devs[i] = tmp + 'Off'
        else:
            devs[i] = 'On'

    if z_scale == 'log':
        log = True
    else:
        log = False

    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap(vals, acts, devs, ax=ax, log=log, cbarlabel=cbarlabel)
    
    texts = annotate_heatmap(im, log=log, textcolors=("white", "black"), valfmt="{x:.0f}")
    
    # create grid for heatmap into every pair
    tcks = np.arange((vals.shape[1])/2)*2 + 1.5
    ax.set_xticks(tcks, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=2)
    ax.tick_params(which="minor", bottom=False, left=False) 
    
    ax.set_title(title)
    fig.tight_layout()
    return fig

def hist_activities(y, scale=None):
    """
    Parameters
    ----------
    y: np array
        label of strings
    scale: None or log
    """
    title = 'Label occurence'
    xlabel = 'counts'
    
    ser = pd.Series(data=y).value_counts()
    ser = ser.sort_values(ascending=True)

    fig, ax = plt.subplots(1, 1, figsize=(9, 3))
    if scale == 'log': 
        plt.xscale('log')
        xlabel = 'log ' + xlabel
    ax.barh(ser.index, ser.values, orientation='horizontal')
    
    plt.xlabel(xlabel)
    fig.suptitle(title)


def corr_devices_01(rep, figsize=(19,14)):
    """ correlation between the on and off states of every device.
    Parameters
    ----------
    rep: pd.DataFrame
        A dataframe where columns are devices and rows are a binary representation
        e.g raw, changepoint, lastfired representation
    
    """
    df = rep.iloc[:,:-1] # get raw without activities
    df = df.reset_index(drop=True)
    df = df.astype(int)

    for device in df.columns:
        mask1 = df[[device]] == 1
        col_off = device + ' Off'
        df[[col_off]] = df[[device]].astype(bool)
        df[[col_off]] = ~df[[col_off]]
        df[[col_off]] = df[[col_off]].astype(int)
        df = df.rename(columns={device: device + ' On'})
        
    dev_count = int(len(df.columns)/2)
    
    ct = df.corr()
    ct_on_on = ct.iloc[:dev_count, :dev_count]
    ct_on_on_values = ct_on_on.values.T
    ct_on_on_devs = list(ct_on_on.index)
    
    ct_on_off = ct.iloc[dev_count:, :dev_count]
    ct_on_off_values = ct_on_off.values.T
    ct_on_off_rows = list(ct_on_off.index)
    ct_on_off_cols = list(ct_on_off.columns)

    fig, (ax1, ax2) = plt.subplots(1,2, figsize=figsize)
    
    # plot on-on corr
    im, cbar = heatmap(ct_on_on_values, ct_on_on_devs, ct_on_on_devs, 
                       ax=ax1, cmap='PuOr', cbarlabel='counts',
                       vmin=-1, vmax=1)

    texts = annotate_heatmap(im, textcolors=("black", "white"), valfmt="{x:.2f}")
    
    # plot on-off corr
    im, cbar = heatmap(ct_on_off_values, ct_on_off_rows, ct_on_off_cols, 
                       ax=ax2, cmap='PuOr', cbarlabel='counts',
                       vmin=-1, vmax=1)
    
    texts = annotate_heatmap(im, textcolors=("black", "white"), valfmt="{x:.2f}")
    
    ax1.set_title("Correlation of on-on signals")
    ax2.set_title("Correlation of on-off signals")
    fig.tight_layout()
    plt.show()

def heatmap_cross_correlation(df_dev, figsize=(10,8)):
    ct = cross_correlation(df_dev)
    vals = ct.values.T.astype(np.float64)
    devs = list(ct.index)
    
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(vals, devs, devs, ax=ax, cmap='BrBG', cbarlabel='counts',
                       vmin=-1, vmax=1)

    texts = annotate_heatmap(im, textcolors=("black", "white"), valfmt="{x:.2f}",
                            threshold=0.6)

    ax.set_title("Cross-correlation of signals")
    fig.tight_layout()
    plt.show()


import pandas as pd

def add_devices_timeint(df_dev, freq='2h'):
    """ creates devices for the specified frequency with value on during
        that frequency
        
    Parameters
    ----------
    df_dev: pd.DataFrame
        devices dataframe, columns start_time, end_time, val, device name
        
    Returns
    ----------
    pd.DataFrame
    """
    df_dev = df_dev.copy()
    # specify 
    st = df_dev['start_time'].iloc[0].ceil('h')
    et = df_dev['end_time'].iloc[-1].floor('h')
    ts=pd.date_range(st, et, freq=freq)
    
    # create Dataframe
    res = pd.DataFrame({'start_time' : ts})
    res['end_time'] = res['start_time'].shift(-1)
    res = res.iloc[:-1,:]
    res['val'] = True
    
    # correct so that no overlap in time is there
    res['start_time'] = res['start_time'] + pd.Timedelta('1ms')
    res['end_time'] = res['end_time'] - pd.Timedelta('1ms')
    
    # create activity label
    res['device'] = res['start_time'].dt.floor('h').dt.strftime("%H:%M:%S")
    res['device2'] = res['end_time'].dt.ceil('h').dt.strftime("%H:%M:%S")
    res['device'] = 'TI: ' + res['device'] + ' - ' + res['device2']
    res = res.iloc[:, :-1]
    
    # combine with dataframe
    res = pd.concat([df_dev, res]).sort_values(by=['start_time'])
    return res.reset_index(drop=True)


def add_devices_dayofweek(df_dev):
    """ creates devices for the specified frequency with value on during
        that frequency
        
    Parameters
    ----------
    df_dev: pd.DataFrame
        devices dataframe, columns start_time, end_time, val, device name
        
    Returns
    ----------
    pd.DataFrame
    """
    df_dev = df_dev.copy()
    # specify 
    
    st = df_dev['start_time'].iloc[0].ceil('d')
    et = df_dev['end_time'].iloc[-1].floor('d')
    ts=pd.date_range(st, et, freq='D').to_series()
    
    #ts.dt.
    
    # create DataFrame
    res = pd.DataFrame({'start_time' : ts})
    res['end_time'] = res['start_time'].shift(-1)
    res = res.iloc[:-1,:]
    res['val'] = True
    
    # correct so that no overlap in time is there
    res['start_time'] = res['start_time'] + pd.Timedelta('1ms')
    res['end_time'] = res['end_time'] - pd.Timedelta('1ms')
    
    # create activity label
    #res['device'] = res['start_time'].dt.floor('h').dt.dayofweek
    res['device'] = res['start_time'].dt.floor('h').dt.day_name()
    #res['device'] = 'WD: ' + res['device'].astype(str)
    
    
    # combine with dataframe
    res = pd.concat([df_dev, res]).sort_values(by=['start_time'])
    return res.reset_index(drop=True)


import pandas as pd
from pyadlml.dataset.activities import START_TIME, END_TIME, ACTIVITY, correct_activities
from pyadlml.dataset.devices import DEVICE, correct_devices
from pyadlml.dataset.obj import Data

def fix_OrdonezB_ADLS(path_to_file):
    """ fixes inconsistent use of tabs for delimiter in the file
    Parameters
    ----------
    path_to_file : str
        path to the file OrdonezB_ADLs.csv
    """
    
    path_corrected = path_to_file[:-17] + 'OrdonezB_ADLs_corr.txt'
    
    with open(path_to_file, 'r') as f_o, open(path_corrected, 'w') as f_t:
        for i, line in enumerate(f_o.readlines()):            
            if i in [0,1]: 
                f_t.write(line)  
                continue
            s = line.split()
            assert len(s) == 5
            new_line = s[0]+' '+s[1]+'\t\t'+s[2]+' '+s[3]+'\t\t'+s[4]                        
            f_t.write(new_line + "\n")
        f_t.close()
        f_o.close()

def _load_activities(act_path):
    df_act = pd.read_csv(act_path, delimiter='\t+', skiprows=[0,1], 
                         names=[START_TIME, END_TIME, ACTIVITY], engine='python')
    df_act[START_TIME] = pd.to_datetime(df_act[START_TIME])
    df_act[END_TIME] = pd.to_datetime(df_act[END_TIME])
    return df_act

def _load_devices(dev_path):
    df_dev = pd.read_csv(dev_path, delimiter='\t+', skiprows=[0, 1], 
                         names=[START_TIME, END_TIME, 'Location', 'Type', 'Place'], 
                         engine='python')
    df_dev[DEVICE] = df_dev['Place'] + ' ' +  df_dev['Location'] + ' ' + df_dev['Type']
    
    # get room mapping devices
    df_locs = df_dev.copy().groupby([DEVICE, 'Type', 'Place', 'Location']).sum()
    df_locs = df_locs.reset_index().drop([START_TIME, END_TIME], axis=1)

    df_dev = df_dev[[START_TIME, END_TIME, DEVICE]]
    df_dev[START_TIME] = pd.to_datetime(df_dev[START_TIME])
    df_dev[END_TIME] = pd.to_datetime(df_dev[END_TIME])
    return df_dev, df_locs

def load(dev_path, act_path, subject):
    """
    """
    assert subject in ['OrdonezA', 'OrdonezB']
    df_act = _load_activities(act_path)
    df_dev, df_loc = _load_devices(dev_path)

    if subject == 'OrdonezB':
        # the activity grooming is often overlapped by sleeping
        # as I deem this activity as important i make it dominant
        
        df_act, cor_lst = correct_activities(df_act, excepts=['Grooming'])
    elif subject == 'OrdonezA':
        df_act, cor_lst = correct_activities(df_act)

    df_dev = correct_devices(df_dev)
    data = Data(df_act, df_dev)
    data.correction_activities = cor_lst
    data.df_dev_rooms = df_loc
    
    return data


from hassbrain_algorithm.controller import Controller, Dataset
from hassbrain_algorithm.datasets._dataset import DataRep
from hassbrain_algorithm.models.hmm.bhmm import BernoulliHMM

# model names

BHMM = 'bhmm'
BHMMPC = 'bhmm_pc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'

SEC1 ='0.01666min'
SEC2 ='0.03333min'
SEC3 ='0.05min'
SEC6 = '0.1min'
SEC12 = '0.2min'
SEC30 = '0.5min'

MIN1 = '1min'
MIN30 = '30min'
DK = Dataset.HASS

DN_KASTEREN = 'kasteren'
DN_HASS_TESTING2 = 'hass_testing2'
DN_HASS_CHRIS = 'hass_chris'
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
BASE_PATH = '/home/cmeier/code/data/thesis_results'
#MD_BASE_PATH = '/home/cmeier/code/tmp'
MD_BASE_PATH = '/home/cmeier/code/tmp'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

#------------------------
# Configuration

MODEL_CLASS = BHSMM
DATA_NAME = DN_HASS_CHRIS_FINAL
DATA_TYPE = DT_CH
TIME_DIFF = MIN30

#------------------------

DATASET_FILE_NAME = 'dataset_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF + ".joblib"
DATASET_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/datasets/' + DATASET_FILE_NAME

MODEL_NAME = 'model_' + MODEL_CLASS + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF
MODEL_FOLDER_PATH = MD_BASE_PATH + '/' + DATA_NAME + '/models/' + MODEL_NAME
#MODEL_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME +".joblib"
MD_LOSS_FILE_PATH = MODEL_FOLDER_PATH + '/' + MODEL_NAME + '.loss.csv'


MD_BASE_PATH = '/home/cmeier/code/data/thesis_results/hass_chris_final/best_model_selection'
MODEL_FILE_PATH = MD_BASE_PATH + '/' + MODEL_NAME + '/' + MODEL_NAME +".joblib"
def main():

    # set of observations
    ctrl = Controller()
    # todo this is only for testing interpretability
    from scripts.test_model import BHMMTestModel
    from scripts.test_model import BHSMMTestModel
    #hmm_model = BHMMTestModel(ctrl)
    hmm_model = BHSMMTestModel(ctrl)
    ctrl.load_dataset_from_file(DATASET_FILE_PATH)
    ctrl.register_model(hmm_model, MODEL_NAME)
    ctrl.register_benchmark(MODEL_NAME)
    ctrl.init_model_on_dataset(MODEL_NAME)
    #ctrl.register_loss_file_path(MD_LOSS_FILE_PATH, MODEL_NAME)
    ctrl.train_model(MODEL_NAME)
    print(MODEL_FILE_PATH)
    ctrl.save_model(MODEL_FILE_PATH, MODEL_NAME)
    print()

if __name__ == '__main__':
    print('dataset filepath : ', DATASET_FILE_PATH)
    print('model filepath: ', MODEL_FILE_PATH)
    print('trainloss filepath: ', MD_LOSS_FILE_PATH)
    main()
    print('finished')


from pyadlml.dataset._representations.raw import create_raw
from pyadlml.dataset._representations.changepoint import create_changepoint
from pyadlml.dataset._representations.lastfired import create_lastfired
import numpy as np

def create_lagged_raw(df_dev, window_size=10, t_res=None, sample_strat='ffill'):
    """ create a 3D tensor of sliding windows over the raw representation.
    Parameters
    ----------
        df_dev: pd.DataFrame
        df_act: pd.DataFrame
        window_size: int
            how much raw vectors should be considered for the creation of the 2d image
        t_res: String
            how much  time intervals TODO ....
            
    Returns
    -------
        res: np.array 3D (K-window_size x window_size x devices)
        res_label: np.array 1D (K-window_size)
    """
    raw = create_raw(df_dev, t_res=t_res, sample_strat=sample_strat)    
    raw = raw.values
    
    return _image_from_reps(raw, window_size)


def create_lagged_changepoint(df_dev, window_size=10, t_res=None):
    """ create a 3D tensor of sliding windows over the raw representation.
    Parameters
    ----------
        df_dev: pd.DataFrame
        df_act: pd.DataFrame
        window_size: int
            how much raw vectors should be considered for the creation of the 2d image
        t_res: String
            how much  time intervals TODO ....
            
    Returns
    -------
        res: np.array 3D (K-window_size x window_size x devices)
        res_label: np.array 1D (K-window_size)
    """
    cp = create_changepoint(df_dev, t_res=t_res).values
    return _image_from_reps(cp, window_size)


def create_lagged_lastfired(df_dev, window_size=10, t_res=None):
    """ create a 3D tensor of sliding windows over the raw representation.
    Parameters
    ----------
        df_dev: pd.DataFrame
        df_act: pd.DataFrame
        window_size: int
            how much raw vectors should be considered for the creation of the 2d image
        t_res: String
            how much  time intervals TODO ....
            
    Returns
    -------
        res: np.array 3D (K-window_size x window_size x devices)
        res_label: np.array 1D (K-window_size)
    """
    lf = create_lastfired(df_dev, t_res=t_res).values
    return _image_from_reps(lf, window_size)


def _image_from_reps(rep, window_size):
    res = np.zeros(shape=(rep.shape[0]-window_size, window_size, rep.shape[1]), dtype=np.int)
    
    for i in range(res.shape[0]):
        tp = rep[i:i+window_size,:]
        res[i,:,:] = tp
    return res


import pandas as pd

from hassbrain_algorithm.controller import Dataset

BHMM = 'bhmm'
BHMMPC = 'bhmmpc'
BHSMM = 'bhsmm'
BHSMMPC = 'bhsmmpc'
HMM = 'hmm'
HMMPC = 'hmmpc'
HSMM = 'hsmm'
HSMMPC = 'hsmmpc'
TADS = 'tads'

SEC1 ='0.01666min'
SEC6 = '0.1min'
SEC20 = '0.3min'
SEC30 = '0.5min'
MIN1 = '1min'

DK = Dataset.HASS

DN_HASS_CHRIS = 'hass_chris'
DN_HASS_CHRIS_FINAL = 'hass_chris_final'
BASE_PATH = '/home/cmeier/code/data/thesis_results'

DT_RAW = 'raw'
DT_CH = 'changed'
DT_LF = 'last_fired'

#------------------------
# Configuration

MODEL_CLASSES = [BHMM, BHSMM, TADS]
DATA_NAME = DN_HASS_CHRIS_FINAL
DATA_TYPE = DT_RAW
TIME_DIFF = SEC30
BEST_MODEL_SELECTION = True
CLASS_ACC_FILE_PATH = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/class_acts.tex'


def main():
    model_dict = {}
    for m_cls in MODEL_CLASSES:
        m_name = 'model_' + m_cls + '_' + DATA_NAME + '_' + DATA_TYPE + '_' + TIME_DIFF

        if BEST_MODEL_SELECTION:
            m_folder_path = BASE_PATH + '/' + DATA_NAME + '/best_model_selection/' + m_name
        else:
            m_folder_path = BASE_PATH + '/' + DATA_NAME + '/models/' + m_name

        model_dict[m_cls] = m_folder_path + '/' + m_name + '.class_acts.csv'


    df = pd.read_csv(next(iter(model_dict.values())))
    unamed0 = 'Unnamed: 0'
    model_col_name = 'Model'
    class_acc_name = 'class acc'
    df.set_index(unamed0)
    df.rename(columns={unamed0:model_col_name}, inplace=True)
    df[model_col_name].iloc[0] = next(iter(model_dict.keys()))
    num_classes = len(df.columns)-1
    df[class_acc_name] = df.iloc[0].drop(model_col_name).sum()/num_classes
    #df[class_acc_name] = [df[]]

    for i, md_key in enumerate(MODEL_CLASSES[1:]):
        tmpdf = pd.read_csv(model_dict[md_key])
        tmpdf.set_index(unamed0)
        tmpdf.rename(columns={unamed0:model_col_name}, inplace=True)
        tmpdf[class_acc_name] = tmpdf.iloc[0].drop(model_col_name).sum()/num_classes

        df = pd.concat([df, tmpdf])
        df[model_col_name].iloc[i+1] = md_key

    df.set_index(model_col_name, inplace=True)
    df = df.transpose()
    texcode = df.to_latex(index=True, float_format=lambda x: '%.3f' % x)

    with open(CLASS_ACC_FILE_PATH, "w") as f:
        f.write(texcode)
        f.close()



if __name__== "__main__":
  main()




import pandas as pd 
from pyadlml.dataset._dataset import label_data
from pyadlml.util import get_npartitions, get_parallel
import dask.dataframe as dd
from pyadlml.dataset import create_raw, START_TIME, END_TIME, TIME, DEVICE



def contingency_table_triggers_01(df_devs, df_acts, idle=False):

    """
    output: 
        read like this: dev 1 turned 123 times from 1 to 0 while act 1 was present
        example
        ---------------------------
                | act 1 | .... | act n|
        ------------------------------- 
        dev 1 0 | 123   |      | 123  | 
        dev 1 1 | 122   |      | 141  |
        ... 
        dev n 0 | 123   |      | 123  | 
        dev n 1 | 122   |      | 141  |
    """
    df = label_data(df_devs, df_acts, idle=idle)
    df['val2'] = df['val'].astype(int)
    return pd.pivot_table(df, 
               columns='activity',
               index=['device', 'val'],
               values='val2',
               aggfunc=len,
               fill_value=0)


def contingency_table_triggers(df_devs, df_acts, idle=False):
    """
    output: 
        read like this: dev 1 was 123 times triggered while act 1 was present
        example
        ---------------------------
                | act 1 | .... | act n|
        ------------------------------- 
        dev 1 | 123   |      | 123  |
        ... 
        dev n | 123   |      | 123  | 
    """
    df = label_data(df_devs, df_acts, idle=idle)    
    df['val'] = 1
    return pd.pivot_table(df, 
               columns='activity',
               index='device',
               values='val',
               aggfunc=len,
               fill_value=0)

def _join_to_interval(x):
    #print('x: ', x)
    return pd.Interval(x[0], x[1])

def _create_cont_df():
    """
    returns: empty pd Dataframe 
        | i | device | val |  activity | time_diff
        -----------------------------
          1 | dev1   | 0   | act1      | td
    """
    df = pd.DataFrame(columns=['device', 'val', 'activity', 'time_diff'])
    df['time_diff'] = pd.to_timedelta(df['time_diff'])
    return df 

def _calc_intervall_diff(i1, i2):
    """
    computes the difference of time between two overlapping intervals
        form new intervall by taking 
        #  |...| 
        #  ~  ~
        # |...|
    
    returns
    -------
        time difference 
    """
    assert i1.overlaps(i2)
    mi = max([i1.left, i2.left])
    ma = min([i1.right, i2.right])
    #return pd.Interval(mi, ma)
    return ma-mi

def contingency_intervals(df_dev, df_act):
    """ compute the crosscorelation by comparing for every interval the binary values
    between the devices
    
    Parameters
    ----------
    df_dev: pd.DataFrame
        device representation 1 
        columns [time, device, val]
    df_act : pd.DataFrame
        contains activities
        columns [start_time, end_time, activity]
        
    Returns
    -------
    output : pd.DataFrame
        read like this: dev 1 turned 123 times from 1 to 0 while act 1 was present
        example
        ---------------------------
                | act 1 | .... | act n|
        ------------------------------- 
        dev 1 0 | 123   |      | 123  | 
        dev 1 1 | 122   |      | 141  |
        ... 
        dev n 0 | 123   |      | 123  | 
        dev n 1 | 122   |      | 141  
    """
    TD = 'time_difference_to_succ'
    
    def func(row, raw, dev_lst):
        """
        """        
        # get selection of relevant devices
        raw_sel = raw[(row.start_time <= raw['time']) & (raw['time'] <= row.end_time)].copy()
        if raw_sel.empty:
            return pd.Series(index=row.index, name=row.name, dtype=row.dtype)
            
        try:
            idx_first = raw_sel.index[0]-1
            raw_sel = raw_sel.append(raw.iloc[idx_first]).sort_values(by=[TIME])
        except:
            print('rs: ', row.start_time)
            print('re: ', row.end_time)
            print('rawsel: ', raw_sel)
            raise ValueError
        
        
        # determine end and start time and correct for the intervals
        raw_start = raw_sel.iloc[0]
        raw_end = raw_sel.iloc[-1]
        t_diff_start = row.start_time - raw_start.time
        t_diff_end = row.end_time - raw_end.time
        raw_sel.at[raw_sel.iloc[0].name, TD] -= t_diff_start
        raw_sel.at[raw_sel.iloc[-1].name, TD] -= t_diff_end
        
        for dev in dev_lst:
            ser = raw_sel.groupby(by=[dev])[TD].sum()
            try:
                dev_on_time = ser.ON
            except:
                dev_on_time = pd.Timedelta('0ns')
            try:
                dev_off_time = ser.OFF
            except:
                dev_off_time = pd.Timedelta('0ns')
                    
            row.at[ser.index.name + " On"] = dev_on_time
            row.at[ser.index.name + " Off"] = dev_off_time        
        return row
    
    def create_meta(raw):
        devices = {name : 'object' for name in raw.columns[1:-1]}
        return {**{'time': 'datetime64[ns]', 'td': 'timedelta64[ns]'}, **devices}
        
    dev_lst = df_dev['device'].unique()
    df_dev = df_dev.sort_values(by='time')
    raw = create_raw(df_dev).applymap(lambda x: 'ON' if x else 'OFF').reset_index()
    raw[TD] = raw['time'].shift(-1) - raw['time']
    
    y = [(d1 + ' Off',d2 + ' On') for d1,d2 in zip(dev_lst, dev_lst)]
    new_cols = [d for tup in y for d in tup] 
            
    
    df_act = df_act.copy().join(pd.DataFrame(index=df_act.index, columns=new_cols))
    
    if not get_parallel():
        df = df_act.apply(func, args=[raw, dev_lst], axis=1)
        df = df.drop(columns=['start_time', 'end_time'])
        df = df.groupby('activity').sum()
        return df.T
    else:
        df = dd.from_pandas(df_act.copy(), npartitions=get_npartitions())\
                .apply(func, args=[raw, dev_lst], axis=1)\
                .drop(columns=['start_time', 'end_time'])\
                .groupby('activity').sum()\
                .compute(scheduler='processes')
                
        return df.T
    



import matplotlib.pyplot as plt
import matplotlib
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from pyadlml.dataset.stats import contingency_intervals, contingency_table_triggers, contingency_table_triggers_01
from pyadlml.dataset.plot.util import func_formatter_sec, heatmap_contingency
from pyadlml.dataset.plot.util import heatmap_square, heatmap, annotate_heatmap
from pyadlml.dataset.activities import add_idle


DEV_CON_HM = {(8,12):(8,7), (8,14):(8,8), (8,3):(12,10), (26,20):(10,10), (11,17):(10,10)}
DEV_CON_01_HM = {(8,14):(12,10), (11,24):(12,6), (11,34):(16,12)}
DEV_CON_01_HM_WT = {(10,24):(14,7), (7, 28):(16,10)}

def _hm_2D_NN(val, hm):
    """ maps nearest neighbor of tuple keys of an hashmap to value
    """
    def dist(i,j):
        return np.sqrt((i[0]-j[0])**2 + (i[1]-j[1])**2)
    
    min_dist = np.inf
    for key in hm:
       diff = dist(key, val)
       if diff < min_dist:
           min_dist = diff
           nearest_neighbor = key
    return hm[nearest_neighbor]

def nk_2_figsize(nk, hm, figsize=None):
    assert isinstance(nk, tuple)
    assert isinstance(hm, dict)
    if figsize is not None:
        return figsize
    else:
        return _hm_2D_NN(nk, hm)

def heatmap_contingency_triggers(df_dev=None, df_act=None, con_tab=None, figsize=None,\
        idle=False, z_scale=None, numbers=True):
    """ computes a table where the device triggers are counted against the activities

    Parameters
    ----------
    df_dev : pd.DataFrame
        device representation 1
    df_activities : pd.DataFrame
        activities of daily living
    con_tab : pd.DataFrame
        the computed contingency table
    Returns
    -------
    fig 
    """
    assert (df_dev is not None and df_act is not None) or con_tab is not None

    title = 'Triggers'
    cbarlabel = 'counts'

    if con_tab is None:
        ct = contingency_table_triggers(df_dev, df_act, idle=idle)
    else:
        ct = con_tab
    
    vals = ct.values.T
    acts = ct.columns
    devs = ct.index
    figsize =  nk_2_figsize(nk=(len(acts), len(devs)), hm=DEV_CON_HM, figsize=figsize)

    fig, ax = plt.subplots(figsize=figsize)
    if z_scale == 'log':
        log = True
    else:
        log = False


    im, cbar = heatmap(vals, acts, devs, ax=ax, log=log, cbarlabel=cbarlabel)

    if numbers:
        texts = annotate_heatmap(im, textcolors=("white", "black"), log=log, valfmt="{x}")

    ax.set_title(title)
    fig.tight_layout()
    plt.show()

def heatmap_contingency_triggers_01(df_dev=None, df_act=None, con_tab_01=None, figsize=None, \
    idle=True, z_scale=None, numbers=True):
    """ computes a table where the device on and off triggers are counted against the activities

    Parameters
    ----------
    df_dev : pd.DataFrame
        device representation 1
    df_activities : pd.DataFrame
        activities of daily living
    con_tab : pd.DataFrame
        the computed contingency table
    Returns
    -------
    fig 
    """
    assert (df_dev is not None and df_act is not None) or con_tab_01 is not None

    title = 'On/Off triggers'
    cbarlabel = 'counts'

    if con_tab_01 is None:
        df_con = contingency_table_triggers_01(df_dev, df_act, idle=idle)
    else:
        df_con = con_tab_01

    tmp = df_con.reset_index()
    tmp['index'] = tmp['device'] + ' ' + tmp['val'].astype(str)
    tmp = tmp.set_index('index')
    tmp = tmp.drop(['device', 'val'], axis=1)

    vals = tmp.values.T
    acts = tmp.columns
    devs = list(tmp.index)

    figsize =  nk_2_figsize(nk=(len(acts), len(devs)), hm=DEV_CON_01_HM, figsize=figsize)

    # replace every second label with 'On'
    for i in range(0,len(devs)):
        if i % 2 != 0:
            devs[i] = 'On'

    if z_scale == 'log':
        log = True
    else:
        log = False



    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap(vals, acts, devs, ax=ax, log=log, cbarlabel=cbarlabel)
    if numbers: 
        texts = annotate_heatmap(im, textcolors=("white", "black"), log=log, valfmt="{x}")
    
    # create grid for heatmap into every pair
    tcks = np.arange((vals.shape[1])/2)*2 + 1.5
    ax.set_xticks(tcks, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=1)
    ax.tick_params(which="minor", bottom=False, left=False) 
    
    ax.set_title(title)
    fig.tight_layout()
    return fig

def heatmap_contingency_overlaps(df_dev=None, df_act=None, con_tab=None, figsize=None, \
    z_scale='log', idle=False, numbers=True):
    """ computes a table where the device on and off intervals are measured against 
    the activities

    Parameters
    ----------
    df_dev : pd.DataFrame
        device representation 1
    df_activities : pd.DataFrame
        activities of daily living
    con_tab : pd.DataFrame
        the computed contingency table
    Returns
    -------
    fig 
    """    
    assert (df_dev is not None and df_act is not None) or con_tab is not None

    title='Cross correlation activites'
    cbarlabel='interval overlap in seconds'

    if con_tab is None:
        if idle:
            df_act = add_idle(df_act.copy())
        df_con = contingency_intervals(df_dev, df_act) 
    else:
        df_con = con_tab

    # convert time to seconds
    df_con = df_con.astype(int)/1000000000
            
    vals = df_con.values.T
    acts = df_con.columns
    devs = list(df_con.index)

    figsize =  nk_2_figsize(nk=(len(acts), len(devs)), hm=DEV_CON_01_HM_WT, figsize=figsize)
    #if z_scale == 'log':
    #    format_func = lambda x, p: func_formatter_sec(np.exp(x), p)
    #else:
    format_func = lambda x, p: func_formatter_sec(x, p)
        
    valfmt = matplotlib.ticker.FuncFormatter(format_func)
    
    heatmap_contingency(vals, acts, devs, cbarlabel, title, 
                         valfmt, figsize, z_scale=z_scale, 
                         numbers=numbers)



import pandas as pd
from pyadlml.dataset import TIME, DEVICE, VAL
from pyadlml.dataset._representations.changepoint import _apply_changepoint
from pyadlml.dataset._dataset import label_data

def create_lastfired(df_devices, t_res=None):
    dev = df_devices.copy()
    lf = _apply_changepoint(dev)
    
    if t_res is not None:
        resampler = lf.resample(t_res, kind='timestamp')
        lf = resampler.apply(_lf_evaluator, df=lf.copy())
        lf = lf.fillna(method='ffill')

    return lf


def _lf_evaluator(series: pd.Series, df):
    """
    is called columnwise for each element of the 
    
    Parameters
    ----------
        series: pd.Series
            conatins name of the column the evaluator operates on
            contains the timestamps for that change in set frame and the value 
            that the specified column has at set timestamp.
            
            is empty if all columns that have no changing entity
            
        df: pd.DataFrame
            contains device representation 3. Is for determining the states of other
            device/columns for the timestamps
            
    Returns: 0|1
    """
    #print('series: ', series, '\n')
    if series.empty:
        return None
    else:         
        # check if this device is triggered during the timeslice
        # this is okay because other devices where triggered -> it has to be zero
        if series.sum() == 0:
            return 0
        
        # get dataframe for timeslice
        ts = series.index[0]
        if len(series) == 1:
            te = ts
        else:
            te = series.index[-1]
        df_slice = df.loc[ts:te]
        
        # if the device changed the state check if it was the 
        # last device to do so in the interval. When true 
        # the timeslice should be 1 for device last fired and 0 otherwise
        for i in range(len(df_slice)-1,-1,-1):
            
            # check if any device triggered
            if df_slice.iloc[i].values.sum() == 0:
                continue
                
            # check if our device was the one who triggered
            elif df_slice.loc[df_slice.index[i], series.name] == 1:
                return 1
            else:
                return 0